/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Originally written by Art Vodolazsky <burivuh@crew.org.ru>
 *  Changed by Andrew Crew Kuznetsov <andrewcrew@rambler.ru>
 */

#include <X11/keysym.h>
#include <string.h>

#include "types.h"
#include "error.h"
#include "xLibrary.h"
#include "xKeyLogManager.h"

xKeyLogBuffer *
xKeyLogBuffer_init ()
{
	xKeyLogBuffer *toInit = NULL;

	toInit = (xKeyLogBuffer *) malloc (sizeof (xKeyLogBuffer));
	if (NULL == toInit)
		return NULL;

	toInit->buffer = xString_init (MAX_WORD_LENGTH);
	if (NULL == toInit->buffer) {
		free (toInit);
		return NULL;
	}
	toInit->iterator = 0;
	toInit->mode = KLB_INSERT_MODE;

	return toInit;
}

status
xKeyLogBuffer_free (xKeyLogBuffer * toFree)
{
	status err = SUCCESS;

	if (NULL != toFree) {
		if (NULL != toFree->buffer)
			free (toFree->buffer);
		else {
			err = WARNING;
		}
		free (toFree);
	}
	else
		err = WARNING;

	return err;
}

status
xKeyLogBuffer_clear (xKeyLogBuffer * klb)
{
	xString_setContent (klb->buffer, NULL);
	klb->iterator = 0;

	return SUCCESS;
}

status
xKeyLogManager_getBufferAction (xEvent * xevent)
{
	status action;
	KeySym keysym = xEvent_getCurrentKeySym (xevent);

	//Cursor keys
	if (IsCursorKey (keysym))
		switch (keysym) {
			case XK_Right:
				action = KLB_MOVE_RIGHT;
				break;
			case XK_Left:
				action = KLB_MOVE_LEFT;
				break;
			default:
				action = KLB_CLEAR;
				break;
		}
	else
		//KeyPad keys
	if (IsKeypadKey (keysym))
		switch (keysym) {
			case XK_KP_Right:
				action = KLB_MOVE_RIGHT;
				break;
			case XK_KP_Left:
				action = KLB_MOVE_LEFT;
				break;
			case XK_KP_Space:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_KP_Delete:
				action = KLB_DEL_SYM_RIGHT;
				break;
			case XK_KP_Insert:
				action = KLB_SWITCH_MODE;
				break;
			default:
				action = KLB_CLEAR;
				break;
		}
	else
		//Func, Mod, PF, PrivateKeypad keys
	if (IsFunctionKey (keysym) || IsModifierKey (keysym) ||
		    IsPFKey (keysym) || IsPrivateKeypadKey (keysym))
		action = KLB_NO_ACTION;
	else
		//MiscFunc keys
	if (IsMiscFunctionKey (keysym))
		if (keysym == XK_Insert) {
			action = KLB_SWITCH_MODE;
		}
		else
			action = KLB_NO_ACTION;
	else
		//del, bkspace, tab, space, return, alpha & num keys
		switch (keysym) {
			case XK_ISO_Prev_Group:
				action = KLB_FLUSH;
				break;
			case XK_comma:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_period:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_Delete:
				action = KLB_DEL_SYM_RIGHT;
				break;
			case XK_BackSpace:
				action = KLB_DEL_SYM_LEFT;
				break;
			case XK_Tab:
				action = KLB_CLEAR;
				break;
			case XK_space:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_Return:
				action = KLB_CLEAR;
				break;
			case XK_Pause:
				action = KLB_NO_ACTION;
				break;
			case XK_Escape:
				action = KLB_NO_ACTION;
				break;
			case XK_Sys_Req:
				action = KLB_NO_ACTION;
				break;
			case XK_exclam:
				action = KLB_SWITCH_LANG; //KLB_FLUSH; KLB_ADD_SYM; //
				break;
			case XK_at:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_numbersign:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_ssharp:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_dollar:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_percent:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_asciicircum:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_ampersand:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_asterisk:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_parenleft:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_parenright:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_hyphen:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_underscore:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_equal:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_plus:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_minus:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_slash:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_bar:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_backslash:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			case XK_question:
				action = KLB_SWITCH_LANG; //KLB_FLUSH;
				break;
			default:
				action = KLB_ADD_SYM;
				send_error (DEBUG, "Key pressed %d\n", &keysym);
				break;
		}

	return action;
}

status
xKeyLogManager_getBindAction (xEvent * xevent, xBindingTable * xbtable)
{
	status action;
	KeySym keysym = xEvent_getCurrentKeySym (xevent);
	KeySym modifier = xEvent_getCurrentModifier (xevent);

	action = xBindingTable_findBindAction (xbtable, keysym, modifier);
	if (action == -1)
		return BA_NONE;
	else
		return action;
}

status
xKeyLogManager_flushBuffer (xKeyLogBuffer * klb, xString * xstring)
{
	char *str = (char *) malloc (MAX_WORD_LENGTH);
	str = memset (str, NULLSYM, MAX_WORD_LENGTH);
	if (klb->iterator != 0) {
		str = memcpy (str, klb->buffer->content, (klb->iterator));
		xString_setContent (xstring, str);
	}
	free (str);
	return SUCCESS;
}

char *
xKeyLogManager_getStringFromEventQueue (xKeyLogBuffer * klb,
					xEvent * xevent,
					xIFocus * xifocus,
					xString * xword,
					xString * xstring,
					xSwitchLang * xswlang, xBindingTable * xbtable)
{
	status action = KLB_NO_ACTION, err = SUCCESS;
	num itr;
	char sym;
	bool changed = FALSE;
	int i;

	xString_setContent (xword, NULL);

	if (xIFocus_hasChanged (xifocus)) {
		xKeyLogBuffer_clear (klb);
		xString_setContent (xstring, NULL);
	}	
	
	/*while (xEvent_checkMousePress (xevent)) {
		printf("Mouse Press\n");
	}*/	
	
	while (xEvent_checkKeyPress (xevent)) {
		if (!xEvent_isNative (xevent)) {
			changed = FALSE;
			continue;
		}

		action = xKeyLogManager_getBindAction (xevent, xbtable);
		xBindingTable_setCurrentAction (xbtable, action);
		if (action > 0) {	//have to perform some bind actions. Flush current buffer and return
			xKeyLogManager_flushBuffer (klb, xword);
			return xstring->content;
		}

		action = xKeyLogManager_getBufferAction (xevent);
		switch (action) {
			case KLB_CLEAR:
				xKeyLogBuffer_clear (klb);
				xString_setContent (xstring, NULL);
				send_error (DEBUG,
 "KLB_CLEAR\n", NULL);
				break;
			case KLB_FLUSH:
				sym = xEvent_getCurrentChar (xevent);
				err = xString_addCharByIndex (klb->buffer, sym, klb->iterator);
				xString_catContent (klb->buffer, xstring);
				xKeyLogBuffer_clear (klb);
				send_error (DEBUG, "KLB_FLUSH\n", NULL);
				send_error (DEBUG, "Current string=%s\n", &xstring->content);
				break;
			case KLB_SWITCH_MODE:
				if (KLB_INSERT_MODE == klb->mode)
					klb->mode = KLB_REPLACE_MODE;
				else
					klb->mode = KLB_INSERT_MODE;
				send_error (DEBUG, "KLB_SWITCH_MODE\n", NULL);
				break;
			case KLB_MOVE_LEFT:
				if (klb->iterator > 0)
					klb->iterator--;
				else {
					xKeyLogBuffer_clear (klb);
					xString_setContent (xstring, NULL);
				}
				send_error (DEBUG, "KLB_MOVE_LEFT\n", NULL);
				break;
			case KLB_MOVE_RIGHT:
				if (klb->iterator < (MAX_WORD_LENGTH - 1))
					klb->iterator++;
				else {
					xKeyLogBuffer_clear (klb);
					xString_setContent (xstring, NULL);
				}
				if (NULLSYM == xString_getCharByIndex (klb->buffer, klb->iterator))
					xKeyLogBuffer_clear (klb);
				send_error (DEBUG, "KLB_MOVE_RIGHT\n", NULL);
				break;
			case KLB_DEL_SYM_LEFT:
				if (klb->iterator > 0) {
					for (itr = klb->iterator;
					     itr < strlen (klb->buffer->content); itr++)
						klb->buffer->content[itr - 1] =
							klb->buffer->content[itr];
					klb->iterator--;
				}
				else {
					xKeyLogBuffer_clear (klb);
					xString_setContent (xstring, NULL);
				}
				send_error (DEBUG, "KLB_DEL_SYM_LEFT\n", NULL);
				break;
			case KLB_DEL_SYM_RIGHT:
				if (strlen (klb->buffer->content) < (MAX_WORD_LENGTH - 1))
					for (itr = klb->iterator;
					     itr < strlen (klb->buffer->content); itr++)
						klb->buffer->content[itr] =
							klb->buffer->content[itr + 1];
				else {
					xKeyLogBuffer_clear (klb);
					xString_setContent (xstring, NULL);
				}
				send_error (DEBUG, "KLB_DEL_SYM_RIGHT\n", NULL);
				break;
			case KLB_ADD_SYM:
				sym = xEvent_getCurrentChar (xevent);
				printf("Add Symbol %c\n", sym);			
				if (klb->mode == KLB_INSERT_MODE) {
					err = xString_addCharByIndex (klb->buffer, sym,
								      klb->iterator);
				}
				else
					err = xString_setCharByIndex (klb->buffer, sym,
								      klb->iterator - 1);
				if (err != SUCCESS) {
					send_error (ERROR, "Index is out of range!\n", NULL);
					break;
				}
				klb->iterator++;
				xswlang->CurrentLang = xEvent_getCurrentLang (xevent);
				//changed = TRUE;
				send_error (DEBUG, "KLB_ADD_SYM\n", NULL);
				
				break;
			case KLB_SWITCH_LANG:
				sym = xEvent_getCurrentChar (xevent);
				if (klb->mode == KLB_INSERT_MODE) {
					err = xString_addCharByIndex (klb->buffer, sym,
								      klb->iterator);
				}
				else
					err = xString_setCharByIndex (klb->buffer, sym,
								      klb->iterator - 1);
				if (err != SUCCESS) {
					send_error (ERROR, "Index is out of range!\n", NULL);
					break;
				}	
				klb->iterator++;			
				//xswlang->CurrentLang = xEvent_getCurrentLang (xevent);	
				changed = TRUE;
				send_error (DEBUG, "KLB_ADD_SYM & KLB_SWITCH_LANG\n", NULL);
				break;
			default:
				send_error (DEBUG, "KLB_NO_ACTION\n", NULL);
				break;
		}
	}

	if (changed) {
		xKeyLogManager_flushBuffer (klb, xword);
                send_error (DEBUG, "Current Word '%s'\n", &(xword->content));
		return xword->content;
	}
	else
		return NULL;
}

status
xKeyLogManager_bind (xBindingTable * xbtable, status key, status modifier, status action)
{
	status err = SUCCESS;
	KeySym key_ks, key_ks_alt, modifier_ks;

	switch (key) {
		case PAUSE_KEY:
			key_ks = XK_Pause;
			key_ks_alt = XK_Break;
			break;
		case SCROLL_LOCK_KEY:
			key_ks = XK_Scroll_Lock;
			key_ks_alt = None;
			break;
		case SYS_RQ_KEY:
			key_ks = XK_Print;
			key_ks_alt = XK_Sys_Req;
			break;
		case None:
			key_ks = None;
			key_ks_alt = None;
			break;
		default:
			send_error (WARNING, "Unknown key to bind\n", NULL);
			err = WARNING;
			break;
	}
	switch (modifier) {
		case SHIFT_KEY:
			modifier_ks = XK_Shift_R;
			break;
		case CONTROL_KEY:
			modifier_ks = XK_Control_R;
			break;
		case None:
			modifier_ks = None;
			break;
		default:
			send_error (WARNING, "Unknown modifier to bind\n", NULL);
			err = WARNING;
			break;
	}

	err = xBindingTable_bindKey (xbtable, key_ks, key_ks_alt, modifier_ks, action);

	return err;
}
