/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Originally written by Art Vodolazsky <burivuh@crew.org.ru>
 */

/* IMPORTANT NOTES:*/
// 1)Work only with null-terminated strings

#include "xLibrary.h"
#include "types.h"
#include "error.h"
#include "utilities.h"

#include <X11/Xmu/Atoms.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <X11/Xdefs.h>
#include <X11/Xmu/Xmu.h>
#include <X11/Intrinsic.h>
#include <X11/Xfuncproto.h>

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

int handleXErrors (Display * d, XErrorEvent * e)
{
	return False;
}

/*working with xObject*/

xObject *xObject_init ()
{
	xObject *toInit = NULL;

	toInit = (xObject *) malloc (sizeof (xObject));
	if (!toInit)
		return NULL;

	toInit->xDisplay = NULL;
	toInit->xScreenNum = WRONG_NUM;
	toInit->xWindow = WRONG_NUM;


	return toInit;
}

Display *xObject_getDisplay (xObject * toGet)
{
	return toGet->xDisplay;
}

status xObject_connect (xObject * client)
{
	client->xDisplay = XOpenDisplay (NULL);
	if (client->xDisplay == NULL) {
		send_error (ERROR, "Can't open default connection to X!\n", NULL);
		return ERROR;
	}
	return SUCCESS;
}

status xObject_setXErrFunc (status (*handler) (Display *, XErrorEvent *))
{
	XSetErrorHandler ((XErrorHandler) handler);
	return SUCCESS;
}

status xObject_createWindow (xObject * toCreate)
{
	toCreate->xWindow = XCreateSimpleWindow (toCreate->xDisplay,
						 DefaultRootWindow (toCreate->
								    xDisplay), 0, 0, 1, 1, 0, 0, 0);
	if (!(toCreate->xWindow)) {
		send_error (ERROR, "Can't create top-level window\n", NULL);
		return ERROR;
	}
	return SUCCESS;
}

Window xObject_getWindow (xObject * toGet)
{
	bool is_destroyed;
	xEvent *xevent = xEvent_init ();
	xevent->communicator->xDisplay = toGet->xDisplay;
	Window win;

	is_destroyed = xEvent_checkDestroyNotify (xevent);
	if (is_destroyed) {
		send_error (WARNING, "Processed window is destroyed!\n", NULL);
		win = ERROR;

	}
	else {
		win = toGet->xWindow;

	}

	xEvent_free (xevent);
	return win;
}

Window xObject_getAppRootWindow (xObject * toGet)
{
	Window backup_window, root_window, parent_window, window = xObject_getWindow (toGet);
	Window *children_window_list;
	unsigned int nchildren;
	bool exit_flag = FALSE;
	if (window == ERROR)
		return ERROR;
	while (!exit_flag) {
		if (!XQueryTree
		    (xObject_getDisplay (toGet), window, &root_window,
		     &parent_window, &children_window_list, &nchildren))
			return 0;
		if (children_window_list)
			XFree (children_window_list);
		if (parent_window == root_window)
			exit_flag = TRUE;
		else {
			backup_window = window;
			window = parent_window;
		}
	}

	return backup_window;
}


Window xObject_findWindowByAtomName (xObject * toFind, char *atom_name)
{
	Atom atom;
	Atom type = None;
	int format;
	unsigned long nitems, after;
	unsigned char *data = NULL;
	Window app_root_window;
	Window tmp_window = 0, win;

	win = xObject_getWindow (toFind);
	if (win == ERROR)
		return ERROR;

	if (win ==
	    RootWindow (xObject_getDisplay (toFind), DefaultScreen (xObject_getDisplay (toFind))))
		return win;

	atom = XInternAtom (toFind->xDisplay, atom_name, True);
	if (!atom)
		return xObject_getWindow (toFind);

	app_root_window = xObject_getAppRootWindow (toFind);
	if (app_root_window == ERROR)
		return ERROR;
	else if (app_root_window == 0)
		return 0;

	XGetWindowProperty (xObject_getDisplay (toFind), app_root_window,
			    atom, 0, 0, False, AnyPropertyType,
			    &type, &format, &nitems, &after, &data);

	if (data)
		XFree (data);

	if (type)
		return app_root_window;

	tmp_window = xObject_findChildByAtom (xObject_getDisplay (toFind), app_root_window, atom);

	if (tmp_window == ERROR)
		return ERROR;
	else if (!tmp_window)
		tmp_window = xObject_getWindow (toFind);

	return tmp_window;
}

Window xObject_findChildByAtom (Display * xDisplay, Window xWindow, Atom atom)
{
	Atom type = None;
	int format;
	unsigned long nitems, after;
	unsigned char *data = NULL;
	Window tmp_window = 0, root_window, parent_window;
	Window *children_window_list;
	unsigned int nchildren;
	unsigned int i;

	if (xWindow <= 0)
		return ERROR;

	if (!XQueryTree (xDisplay, xWindow, &root_window, &parent_window,
			 &children_window_list, &nchildren))
		return 0;

	for (i = 0; !tmp_window && (i < nchildren); i++) {
		data = NULL;

		XGetWindowProperty (xDisplay, children_window_list[i], atom,
				    0, 0, False, AnyPropertyType, &type,
				    &format, &nitems, &after, &data);
		if (data)
			XFree (data);

		if (type)
			tmp_window = children_window_list[i];
	}

	for (i = 0; !tmp_window && (i < nchildren); i++)
		tmp_window = xObject_findChildByAtom (xDisplay, children_window_list[i], atom);

	if (children_window_list)
		XFree (children_window_list);

	return tmp_window;
}

XClassHint *xObject_getWMClass (xObject * toGet)
{
	int dummyI;
	unsigned int dummyU;
	XClassHint *wm_class = NULL;
	Window root_window, window;

	window = xObject_findWindowByAtomName (toGet, "WM_CLASS");
	if (window == ERROR)
		return NULL;
	if (!window)
		window = xObject_getWindow (toGet);

	wm_class = XAllocClassHint ();
	if (!XGetClassHint (xObject_getDisplay (toGet), window, wm_class)) {
		//send_error (ERROR, "Can't get WM_CLASS\n", NULL);
		XFree (wm_class->res_class);
		XFree (wm_class->res_name);
		XFree (wm_class);
		return NULL;
	}
	return wm_class;
}

//NOTE: malloc! Don't forget to free after use
char *xObject_getWMClass_name (xObject * toGet)
{
	char *string = NULL;
	XClassHint *wm_class;

	wm_class = xObject_getWMClass (toGet);
	if (!wm_class)
		return NULL;

	string = malloc (strlen (wm_class->res_name) + 1);

	if (string) {
		string = memset (string, NULLSYM, strlen (wm_class->res_name) + 1);
		string = memcpy (string, wm_class->res_name, strlen (wm_class->res_name));
	}

	XFree (wm_class->res_class);
	XFree (wm_class->res_name);
	XFree (wm_class);

	return string;
}

//NOTE: malloc! Don't forget to free after use
char *xObject_getWMClass_class (xObject * toGet)
{
	char *string = NULL;
	XClassHint *wm_class;

	wm_class = xObject_getWMClass (toGet);
	if (!wm_class)
		return NULL;

	string = malloc (strlen (wm_class->res_class) + 1);

	if (string) {
		string = memset (string, NULLSYM, strlen (wm_class->res_class) + 1);
		string = memcpy (string, wm_class->res_class, strlen (wm_class->res_class));
	}

	XFree (wm_class->res_class);
	XFree (wm_class->res_name);
	XFree (wm_class);

	return string;
}

status xObject_free (xObject * toDelete)
{
	if (toDelete) {
		free (toDelete);
	}
	else
		return WARNING;

	return SUCCESS;
}

status xObject_setEventMask (xObject * toSet, int event_mask) //mask
{
	Window win = xObject_getWindow (toSet);
	status err = SUCCESS;
	if (win != ERROR)
		XSelectInput (xObject_getDisplay (toSet), win, event_mask);
	else
		err = ERROR;

	return err;
}

/*working with xString*/

xString *xString_init (size init_size)
{
	xString *toInit = NULL;

	toInit = (xString *) malloc (sizeof (xString));
	if (!toInit)
		return NULL;

	toInit->content = NULL;

	if (!init_size) {
		toInit->currentSize = WRONG_SIZE;
	}
	else
		toInit->currentSize = init_size;

	if (init_size) {
		toInit->content = (char *) malloc (init_size);
		if (!(toInit->content)) {
			xString_free (toInit);
			return NULL;
		}
		toInit->keycode = (KeyCode *) malloc (sizeof (KeyCode) * init_size);
		if (!(toInit->keycode)) {
			xString_free (toInit);
			return NULL;
		}
		toInit->keycodeModifiers = (mask *) malloc (sizeof (mask) * init_size);
		if (!(toInit->keycodeModifiers)) {
			xString_free (toInit);
			return NULL;
		}
		toInit->home = xObject_init ();
		if (!toInit->home) {
			xString_free (toInit);
			return NULL;
		}

		memset (toInit->content, NULLSYM, init_size);
		memset (toInit->keycode, 0, init_size);
		memset (toInit->keycodeModifiers, 0, init_size);
	}

	return toInit;
}

status xString_free (xString * toDelete)
{
	status err = SUCCESS;

	if (toDelete) {
		if (xObject_free (toDelete->home) != SUCCESS);
		err = WARNING;
		if ((toDelete->keycodeModifiers))
			free ((toDelete->keycodeModifiers));
		else
			err = WARNING;
		if (toDelete->keycode)
			free (toDelete->keycode);
		else
			err = WARNING;
		if (toDelete->content)
			free (toDelete->content);
		else
			err = WARNING;

		free (toDelete);
	}
	else
		err = WARNING;

	return err;
}

status xString_setKeycode (xString * toChange, mask keycodeModifiers)
{
	unsigned int charIndex;
	KeySym ks;
	char element[2] = { NULLSYM, NULLSYM };

	if (xObject_getDisplay (toChange->home) == NULL)
		return INCORRECT_OPERATION;
	if ((WRONG_SIZE == (toChange->currentSize))
	    || (0 == (toChange->currentSize)))
		return INCORRECT_OPERATION;
	toChange->keycodeModifiers =
		memset (toChange->keycodeModifiers, 0, sizeof (mask) * (toChange->currentSize));
	for (charIndex = 0; charIndex < strlen (toChange->content); charIndex++) {
		element[0] = toChange->content[charIndex];
		if (!(element[0]))
			break;
		if (isUpperNonAlphaCyr (element[0]) || (isupper (element[0])))
			toChange->keycodeModifiers[charIndex] = (keycodeModifiers | ShiftMask);
		else
			toChange->keycodeModifiers[charIndex] = (keycodeModifiers | None);

		element[0] = getLowerCyrNonAlpha (element[0]);
		ks = xEvent_CharToKeysym (element[0]);
		toChange->keycode[charIndex] =
			XKeysymToKeycode (xObject_getDisplay (toChange->home), ks);

		if (!ks)
			toChange->keycode[charIndex] =
				XKeysymToKeycode (xObject_getDisplay
						  (toChange->home), XStringToKeysym (element));

	}

	return SUCCESS;
}

status xString_ToASCII (xString * xstring)
{
	char *array = xstring->content;
	size arraySize = WRONG_SIZE;
	num counter = WRONG_NUM, index = WRONG_NUM;
	char element[CYR_BYTES_PER_SYM + 1] = { NULLSYM, NULLSYM, NULLSYM };

	if (!array)
		return NULL_PTR;
	arraySize = strlen (array);
	counter = 0;
	index = 0;

	if (isascii (*(array))) {
		xString_setKeycode (xstring, None);
		return ENGLISH;
	}

	while (index <= (arraySize) / CYR_BYTES_PER_SYM) {
		element[0] = array[counter];
		element[1] = array[counter + 1];

		array[index] = get_ascii (element);
		counter += CYR_BYTES_PER_SYM;
		index++;
	}
	memset ((array + index), NULLSYM, xstring->currentSize - index);
	xString_setKeycode (xstring, CyrMask);

	return RUSSIAN;
}

status xString_setContent (xString * toChange, char *string)
{
	size sizeToChange;
	size sizeOld;
	size fullsizeOld;
	sizeOld = strlen (toChange->content);
	fullsizeOld = toChange->currentSize;

	if (string == NULL) {
		memset (toChange->content, NULLSYM, fullsizeOld);
		return NULL_PTR;
	}

	sizeToChange = strlen (string);

	if (sizeOld == WRONG_SIZE)
		sizeOld = 0;

	if (!sizeToChange) {
		if ((sizeOld > 0)) {
			memset (toChange->content, NULLSYM, sizeOld);
			memset (toChange->keycode, 0, sizeof (KeyCode) * sizeOld);
			memset (toChange->keycodeModifiers, 0, sizeof (mask) * sizeOld);

			return SUCCESS;
		}
		else
			return WARNING;
	}
	else {
		if (sizeToChange >= fullsizeOld) {
			if (!(realloc (toChange->content, sizeToChange + 1))
			    || !(realloc (toChange->keycode, sizeof (KeyCode) * (sizeToChange + 1)))
			    ||
			    !(realloc
			      (toChange->keycodeModifiers, sizeof (mask) * (sizeToChange + 1)))) {
				xString_free (toChange);
				return BAD_ALLOC;
			}
			else {
				memset (toChange->content, NULLSYM, sizeToChange + 1);
				memset (toChange->keycode, 0,
					sizeof (KeyCode) * (sizeToChange + 1));
				memset (toChange->keycodeModifiers, 0,
					sizeof (mask) * (sizeToChange + 1));

				toChange->currentSize = sizeToChange + 1;
			}
		}
		else {		// fullsizeOld > sizeToChange
			memset (toChange->content, NULLSYM, fullsizeOld);
			memset (toChange->keycode, 0, sizeof (KeyCode) * fullsizeOld);
			memset (toChange->keycodeModifiers, 0, sizeof (mask) * fullsizeOld);
		}
		memcpy (toChange->content, string, sizeToChange);
	}

	return SUCCESS;
}

status xString_catContent (xString * src, xString * dest)
{
	num len_src = xString_getLength (src);
	num len_dest = xString_getLength (dest);
	char *tmp;

	if (xString_getLength (src) == 0)
		return NULL_PTR;
	tmp = malloc (len_dest + len_src + 1);
	if (tmp == NULL)
		return ERROR;

	memset (tmp, NULLSYM, len_dest + len_src + 1);
	strncat (tmp, dest->content, len_dest);
	strncat (tmp, src->content, len_src);

	xString_setContent (dest, tmp);

	free (tmp);
}

status xString_addCharByIndex (xString * toChange, char sym, num index)
{
	char *buffer = NULL;

	if (index > toChange->currentSize - 2)
		return ERROR;

//      send_error(DEBUG, "xString_addCharByIndex_began", NULL);

	if (index == (toChange->currentSize - 2)) {
		(toChange->currentSize) += (toChange->currentSize);
		toChange->content =
			(char *) realloc (toChange->content,
					  sizeof (char) * (toChange->currentSize));
		toChange->keycode =
			(KeyCode *) realloc (toChange->keycode,
					     sizeof (KeyCode) * (toChange->currentSize));
		toChange->keycodeModifiers =
			(mask *) realloc (toChange->keycodeModifiers,
					  sizeof (mask) * (toChange->currentSize));
	}

	if (strlen (toChange->content) > index) {
		buffer = (char *) malloc (toChange->currentSize);
		buffer = memcpy (buffer, toChange->content + index,
				 strlen (toChange->content) - index);
	}
	else
		buffer = NULL;

	toChange->content[index] = sym;
	toChange->content[index + 1] = NULLSYM;
	if (buffer != NULL) {
		toChange->content = strcat (toChange->content, buffer);
		free (buffer);
	}

//      send_error(DEBUG, "xString_addCharByIndex_ended", NULL);

	return SUCCESS;
}

status xString_setCharByIndex (xString * toChange, char sym, num index)
{
	if (index >= toChange->currentSize)
		return ERROR;

	toChange->keycode = memset (toChange->keycode, NULLSYM, toChange->currentSize);
	toChange->keycodeModifiers =
		memset (toChange->keycodeModifiers, NULLSYM, toChange->currentSize);

	toChange->content[index] = sym;

	return SUCCESS;
}

char xString_getCharByIndex (xString * toGet, num index)
{
	if (index < toGet->currentSize)
		return toGet->content[index];
	else
		return NULLSYM;
}

size xString_getLength (xString * toGet)
{
	return strlen (toGet->content);
}

/*working with xIFocus*/

xIFocus *xIFocus_init ()
{
	xIFocus *toInit = NULL;

	toInit = (xIFocus *) malloc (sizeof (xIFocus));
	if (!toInit)
		return NULL;
	toInit->owner = xObject_init ();
	toInit->ptr = NULL;
	if (!(toInit->owner)) {
		free (toInit);
		return NULL;
	}
	toInit->ptr_num = 0;
	toInit->cur_status = WRONG_NUM;
	toInit->changed = FALSE;

	return toInit;
}

status xIFocus_free (xIFocus * toDelete)
{
	status err = SUCCESS;

	if (toDelete) {
		if ((xObject_free (toDelete->owner) != SUCCESS))
			err = WARNING;
		while(toDelete->ptr_num > 0){
			xObject_free(toDelete->ptr[--(toDelete->ptr_num)]);
		}
		if(toDelete->ptr)
			free(toDelete->ptr);
		free (toDelete);
	}
	else
		err = WARNING;

	return err;
}

status xIFocus_update (xIFocus * toUpdate)
{
	Window oldwin = xObject_getWindow (toUpdate->owner);
	bool isSameScreen;
	int dummy;
	unsigned int dummyU;
	Window rootW = 0, childW = 0;
	status err;

	if (!xObject_getDisplay (toUpdate->owner))
		return INCORRECT_OPERATION;
	XGetInputFocus (xObject_getDisplay (toUpdate->owner),
			&(toUpdate->owner->xWindow), &(toUpdate->cur_status));
	if (None == toUpdate->owner->xWindow)
		return ERROR;
	
	if (oldwin != xObject_getWindow (toUpdate->owner))
		toUpdate->changed = TRUE;
	else
		toUpdate->changed = FALSE;
	
	while( (toUpdate->ptr_num) > 0){
		xObject_free(toUpdate->ptr[--(toUpdate->ptr_num)]);
	}
	if( (toUpdate->ptr) != NULL){
		free(toUpdate->ptr);
		toUpdate->ptr = NULL;
	}
	
	childW = xObject_getWindow (toUpdate->owner);
	if (childW == ERROR)
		return ERROR;
	
	if ( (toUpdate->cur_status == RevertToPointerRoot) ) {
		if (xObject_getWindow (toUpdate->owner) == ERROR)
			return ERROR;

		isSameScreen =
			XQueryPointer (xObject_getDisplay (toUpdate->owner),
				       childW, &rootW, &childW, &dummy,
				       &dummy, &(dummy), &(dummy), &dummyU);
		if (isSameScreen == TRUE) {
			while (childW != 0) {
				(toUpdate->ptr_num)++;
				toUpdate->ptr = realloc(toUpdate->ptr, sizeof(void*)*(toUpdate->ptr_num));
				toUpdate->ptr[(toUpdate->ptr_num) - 1] = xObject_init();
				
				toUpdate->ptr[(toUpdate->ptr_num) - 1]->xDisplay = xObject_getDisplay(toUpdate->owner);
				toUpdate->ptr[(toUpdate->ptr_num) - 1]->xWindow = childW;
		
				if ( (xObject_getWindow (toUpdate->ptr[(toUpdate->ptr_num) - 1]) ) == ERROR)
					return ERROR;
				isSameScreen =
					XQueryPointer (xObject_getDisplay (toUpdate->owner),
						       childW, &rootW,
						       &childW, &dummy,
						       &dummy, &(dummy), &(dummy), &dummyU);
			}
		}
	}

	return SUCCESS;
}

bool xIFocus_hasChanged (xIFocus * ifocus)
{
	if (ifocus->changed == TRUE)
		return TRUE;
	else
		return FALSE;
}

/*working with xEvent*/

xEvent *xEvent_init ()
{
	xEvent *toInit = NULL;

	toInit = (xEvent *) malloc (sizeof (xEvent));
	if (!toInit)
		return NULL;
	toInit->communicator = xObject_init ();
	if (NULL == toInit->communicator) {
		xObject_free (toInit->communicator);
		free (toInit);
		return NULL;
	}

	return toInit;
}

status xEvent_free (xEvent * toDelete)
{
	status err = SUCCESS;

	if (toDelete) {
		if (xObject_free (toDelete->communicator) != SUCCESS);
		err = WARNING;
		free (toDelete);
	}
	else
		err = WARNING;

	return err;
}

status xEvent_setCommunicator (xEvent * xevent, xObject * toSet)
{
	xevent->communicator = memcpy (xevent->communicator, toSet, sizeof (xObject));
	return SUCCESS;
}

Window xEvent_getCommunicatorWindow (xEvent * xevent)
{
	return xObject_getWindow (xevent->communicator);
}

Window xEvent_getEventWindow (xEvent * xevent)
{
	return xevent->event.xkey.window;
}

status xEvent_sendXKey (xEvent * toSend, KeyCode keycode, mask keycode_modifiers)
{
	status err = SUCCESS;
	Window win;
	
	if (!(xObject_getDisplay (toSend->communicator)))
		return INCORRECT_OPERATION;

	toSend->event.type = KeyPress;
	toSend->event.xkey.type = KeyPress;
	toSend->event.xkey.window = toSend->communicator->xWindow;
	toSend->event.xkey.root =
		RootWindow (xObject_getDisplay (toSend->communicator),
			    DefaultScreen (xObject_getDisplay (toSend->communicator)));
	toSend->event.xkey.subwindow = None;
	toSend->event.xkey.same_screen = True;
	toSend->event.xkey.display = toSend->communicator->xDisplay;
	toSend->event.xkey.state = keycode_modifiers;
	toSend->event.xkey.keycode = keycode;
	toSend->event.xkey.time = CurrentTime;

	win = xObject_getWindow (toSend->communicator);
	if (win == ERROR)
		return ERROR;
	err = XSendEvent (xObject_getDisplay (toSend->communicator), win,
			  TRUE, CLIENT_EVENT_MASK, &(toSend->event));
	if (err) {
		toSend->event.type = KeyRelease;
		toSend->event.xkey.type = KeyRelease;
		toSend->event.xkey.time = CurrentTime;

		win = xObject_getWindow (toSend->communicator);
		if (win == ERROR)
			return ERROR;
		err = XSendEvent (xObject_getDisplay (toSend->communicator),
				  win, TRUE, CLIENT_EVENT_MASK, &(toSend->event));
		if (!err)
			return INCORRECT_OPERATION;
	}
	else
		return INCORRECT_OPERATION;

	return err;
}

status xEvent_sendEvent (xEvent * toSend, XEvent * event, mask event_mask)
{
	status err;

	err = XSendEvent (xObject_getDisplay (toSend->communicator),
			  xObject_getWindow (toSend->communicator), TRUE, event_mask, event);
	if (!err)
		return INCORRECT_OPERATION;
	else
		return SUCCESS;
}

status xEvent_putBackToQueue (xEvent * toPut, XEvent * event)
{
	status err;

	err = XPutBackEvent (xObject_getDisplay (toPut->communicator), event);
	if (!err)
		return INCORRECT_OPERATION;
	else
		return SUCCESS;
}

status xEvent_sendString (xEvent * xevent, xString * xstring)
{
	unsigned int charIndex = 0;
	status err = SUCCESS;

	for (charIndex = 0; charIndex < strlen (xstring->content); charIndex++) {
		err = xEvent_sendXKey (xevent, xstring->keycode[charIndex],
				       xstring->keycodeModifiers[charIndex]);
		if (err == ERROR)
			return ERROR;
	}
	XFlush (xevent->communicator->xDisplay);

	return err;
}

bool xEvent_checkKeyRelease (xEvent * toCheck)
{
	status err;

	err = XCheckMaskEvent (toCheck->communicator->xDisplay, KeyReleaseMask, &toCheck->event);
	if (err)
		return TRUE;
	else
		return FALSE;
}

bool xEvent_checkKeyPress (xEvent * toCheck)
{
	status err;

	err = XCheckMaskEvent (xObject_getDisplay (toCheck->communicator),
			       KeyPressMask, &toCheck->event);
	if (err)
		return TRUE;
	else
		return FALSE;
}

bool xEvent_checkMousePress (xEvent * toCheck)
{
	status err;
	err = XCheckMaskEvent (xObject_getDisplay (toCheck->communicator),
			       ButtonPressMask, &toCheck->event);
	if (err)
		return TRUE;
	else
		return FALSE;
}

bool xEvent_checkDestroyNotify (xEvent * toCheck)
{
	status err;

	err = XCheckMaskEvent (xObject_getDisplay (toCheck->communicator),
			       StructureNotifyMask, &toCheck->event);
	if (err && (toCheck->event.type == DestroyNotify))
		return TRUE;
	else
		return FALSE;
}

bool xEvent_isNative (xEvent * toCheck)
{
	if (toCheck->event.xkey.send_event == TRUE)
		return FALSE;
	else
		return TRUE;
}

char xEvent_getCurrentChar (xEvent * toGet)
{
	KeySym ks;
	int nbytes;
	char str[MAX_WORD_LENGTH];
	str[0] = NULLSYM;
	char *strPtr = NULL;

	nbytes = XLookupString (&(toGet->event.xkey), str, MAX_WORD_LENGTH, &ks, NULL);
	//if ((str[0] == NULLSYM)) {
		ks = xEvent_KeysymToLatin1 (ks);

		str[0] = xEvent_KeysymToChar (ks);
		if ((ks != 0) && (str[0] == NULLSYM)) {
			strPtr = XKeysymToString (ks);
			str[0] = strPtr[0];
		}
	//}
	return str[0];
}

KeySym xEvent_getCurrentKeySym (xEvent * toGet)
{
	KeySym ks;
	int nbytes;
	char *str = (char *) malloc (MAX_WORD_LENGTH);
	nbytes = XLookupString (&(toGet->event.xkey), str, MAX_WORD_LENGTH, &ks, NULL);
	free (str);

	return ks;
}

KeySym xEvent_getCurrentModifier (xEvent * toGet)
{
	KeySym ks;

	if ((toGet->event.xkey.state) & ControlMask)
		ks = XK_Control_R;
	else if ((toGet->event.xkey.state) & ShiftMask)
		ks = XK_Shift_R;
	else
		ks = None;
	return ks;
}

KeySym xEvent_CharToKeysym (char sym)
{
	num iterator;
	KeySym res = None;
	KeySym ks[] = { XK_quotedbl, XK_apostrophe, XK_bracketleft, XK_bracketright,
		XK_comma, XK_period, XK_colon, XK_semicolon,
		XK_less, XK_greater, XK_braceleft, XK_braceright,
		XK_space, XK_exclam, XK_at, XK_numbersign,
		XK_ssharp, XK_dollar, XK_percent, XK_degree,
		XK_ampersand, XK_asterisk, XK_parenleft, XK_parenright,
		XK_underscore, XK_equal, XK_plus, XK_minus,
		XK_slash, XK_bar, XK_backslash, XK_question
	};
	char ch[] = { '"', '\'', '{', '}',
		',', '.', ':', ';',
		'<', '>', '[', ']',
		' ', '!', '@', '#',	//TODO
		'#', '$', '%', '^',	//TODO
		'&', '*', '(', ')',
		'_', '=', '+', '-',
		'/', '|', '\\', '?'
	};

	for (iterator = 0; iterator < 32; iterator++)
		if (sym == ch[iterator]) {
			res = ks[iterator];
			break;
		}
	return res;
}

char xEvent_KeysymToChar (KeySym keysym)
{
	num iterator;
	char res = NULLSYM;
	KeySym ks[] = { XK_quotedbl, XK_apostrophe, XK_bracketleft, XK_bracketright,
		XK_comma, XK_period, XK_colon, XK_semicolon,
		XK_less, XK_greater, XK_braceleft, XK_braceright,
		XK_space, XK_exclam, XK_at, XK_numbersign,
		XK_ssharp, XK_dollar, XK_percent, XK_degree,
		XK_ampersand, XK_asterisk, XK_parenleft, XK_parenright,
		XK_underscore, XK_equal, XK_plus, XK_minus,
		XK_slash, XK_bar, XK_backslash, XK_question
	};
	char ch[] = { '"', '\'', '{', '}',
		',', '.', ':', ';',
		'<', '>', '[', ']',
		' ', '!', '@', '#',	//TODO
		'#', '$', '%', '^',
		'&', '*', '(', ')',
		'_', '=', '+', '-',
		'/', '|', '\\', '?'
	};

	for (iterator = 0; iterator < 32; iterator++)
		if (keysym == ks[iterator]) {
			res = ch[iterator];
			break;
		}
	return res;
}

KeySym xEvent_KeysymToLatin1 (KeySym ks)
{
	KeySym ks_latin1 = 0;
	char *ksname = XKeysymToString (ks);
	num iterator;
	KeySym rus[] = { XK_Cyrillic_E, XK_Cyrillic_e, XK_Cyrillic_HA,
		XK_Cyrillic_HARDSIGN,
		XK_Cyrillic_be, XK_Cyrillic_yu, XK_Cyrillic_ZHE,
		XK_Cyrillic_zhe,
		XK_Cyrillic_BE, XK_Cyrillic_YU, XK_Cyrillic_ha,
		XK_Cyrillic_hardsign,
		XK_Cyrillic_EF, XK_Cyrillic_I, XK_Cyrillic_ES, XK_Cyrillic_VE,
		XK_Cyrillic_U, XK_Cyrillic_A, XK_Cyrillic_PE, XK_Cyrillic_ER,
		XK_Cyrillic_SHA, XK_Cyrillic_O, XK_Cyrillic_EL,
		XK_Cyrillic_DE,
		XK_Cyrillic_SOFTSIGN, XK_Cyrillic_TE, XK_Cyrillic_SHCHA,
		XK_Cyrillic_ZE,
		XK_Cyrillic_SHORTI, XK_Cyrillic_KA, XK_Cyrillic_YERU,
		XK_Cyrillic_IE,
		XK_Cyrillic_GHE, XK_Cyrillic_EM, XK_Cyrillic_TSE,
		XK_Cyrillic_CHE,
		XK_Cyrillic_EN, XK_Cyrillic_YA,
		XK_Cyrillic_ef, XK_Cyrillic_i, XK_Cyrillic_es, XK_Cyrillic_ve,
		XK_Cyrillic_u, XK_Cyrillic_a, XK_Cyrillic_pe, XK_Cyrillic_er,
		XK_Cyrillic_sha, XK_Cyrillic_o, XK_Cyrillic_el,
		XK_Cyrillic_de,
		XK_Cyrillic_softsign, XK_Cyrillic_te, XK_Cyrillic_shcha,
		XK_Cyrillic_ze,
		XK_Cyrillic_shorti, XK_Cyrillic_ka, XK_Cyrillic_yeru,
		XK_Cyrillic_ie,
		XK_Cyrillic_ghe, XK_Cyrillic_em, XK_Cyrillic_tse,
		XK_Cyrillic_che,
		XK_Cyrillic_en, XK_Cyrillic_ya
	};
	KeySym eng[] = { XK_quotedbl, XK_apostrophe, XK_bracketleft, XK_bracketright,
		XK_comma, XK_period, XK_colon, XK_semicolon,
		XK_less, XK_greater, XK_braceleft, XK_braceright,
		XK_A, XK_B, XK_C, XK_D,
		XK_E, XK_F, XK_G, XK_H,
		XK_I, XK_J, XK_K, XK_L,
		XK_M, XK_N, XK_O, XK_P,
		XK_Q, XK_R, XK_S, XK_T,
		XK_U, XK_V, XK_W, XK_X,
		XK_Y, XK_Z,
		XK_a, XK_b, XK_c, XK_d,
		XK_e, XK_f, XK_g, XK_h,
		XK_i, XK_j, XK_k, XK_l,
		XK_m, XK_n, XK_o, XK_p,
		XK_q, XK_r, XK_s, XK_t,
		XK_u, XK_v, XK_w, XK_x,
		XK_y, XK_z
	};
	if (!strncmp (ksname, "Cyrillic", 8)) {
		for (iterator = 0; iterator < 66; iterator++)
			if (ks == rus[iterator]) {
				ks_latin1 = eng[iterator];
				break;
			}
	}
	else {
		//send_error (ERROR, "Language not supported\n", NULL);
		ks_latin1 = ks;
	}

	return ks_latin1;
}

status xEvent_getCurrentLang (xEvent * toGet)
{
	KeySym ks = xEvent_getCurrentKeySym (toGet);
	char *ksname = XKeysymToString (ks);
	send_error (DEBUG, "System is return symbol '%s'!\n", &ksname);
	if (!strncmp (ksname, "Cyrillic", 8))
		return RUSSIAN;
	else
		return ENGLISH;
}

status xEvent_getSelectedText (xEvent * toGet)
{
	XSelectionRequestEvent *req;
	//xObject_setEventMask (toGet->xobject,StructureNotifyMask+ExposureMask);
	//XNextEvent (toGet->requestor->xDisplay, toGet->event);
	if (toGet->event.type == SelectionRequest)
		//
		// Somebody wants our data
		//
	{
		printf ("Sel\n");		
	}
}
/*working with xSelection*/

xSelection *xSelection_init (Display * ManagedDisplay)
{
	xSelection *toInit = NULL;

	toInit = (xSelection *) malloc (sizeof (xSelection));
	if (!toInit)
		return NULL;
	toInit->owner = xObject_init ();
	toInit->requestor = xObject_init ();
	if (!(toInit->owner) || !(toInit->requestor)) {
		xObject_free (toInit->owner);
		xObject_free (toInit->requestor);
		free (toInit);
		return NULL;
	}

	toInit->selection = XA_PRIMARY;
	toInit->property = XInternAtom (ManagedDisplay, "_selected_STRING", FALSE);

	if (!X_HAVE_UTF8_STRING) {
		send_error (ERROR, "XServer doesn't support UTF8\n", NULL);
		return NULL;
	}
	return toInit;
}

status xSelection_free (xSelection * toDelete)
{
	status err = SUCCESS;

	if (toDelete) {
		if ((xObject_free (toDelete->owner) != SUCCESS)
		    || (xObject_free (toDelete->requestor) != SUCCESS))
			err = WARNING;
		free (toDelete);
	}
	else
		err = WARNING;
	return err;
}

char *xSelection_getToString (xSelection * toGet, XSelectionEvent * xselection)
{
	int format, exit_flag;
	unsigned long len, bytes_left = 1, offset, part_len;
	unsigned char *data;
	int result;
	char *string = NULL;
	Atom type;
	Display *ManagedDisplay = xObject_getDisplay (toGet->owner);
	Window ManagedWindow = xObject_getWindow (toGet->owner);

	part_len = 8;
	offset = 0;
	result = XGetWindowProperty (ManagedDisplay, ManagedWindow,
				     xselection->property,
				     offset, part_len,
				     FALSE, AnyPropertyType,
				     &type, &format, &len, &bytes_left, &data);
	if (result == Success) {
		string = (char *) malloc ((format / 8) * (len + bytes_left) + 1);
		string = memset (string, 0, (format / 8) * (len + bytes_left) + 1);
		string = strcat (string, data);
		offset += part_len;
	}
	else
		return NULL;
	while (bytes_left > 0) {
//try to get minimum portion of data from our recieve property to get the size of data (len*(format/8) )
		result = XGetWindowProperty (ManagedDisplay, ManagedWindow,
					     xselection->property,
					     offset, part_len,
					     FALSE, AnyPropertyType,
					     &type, &format, &len, &bytes_left, &data);

		if (result != Success) {
			send_error (ERROR, "Fail to get selected text\n", NULL);
			if (string)
				free (string);
			return NULL;
		}
		else {
			string = strcat (string, data);
			offset += part_len;
		}
		XFree (data);
	}
//We've already finish working with our recieve property. Notify everybody we'll done.
	XDeleteProperty (ManagedDisplay, ManagedWindow, xselection->property);

	return string;
}

status xSelection_chooseValidTarget (xSelection * toGet)
{
	XEvent event;
	char *list;
	num format, i;
	unsigned long len, bytes_left, dummy;
	unsigned char *data;
	Atom type;
	num result;
	status err = SUCCESS;

	Atom *targetP = NULL;
	bool UTF8_S_Enable = FALSE;
	bool S_Enable = FALSE;
	Display *ManagedDisplay = xObject_getDisplay (toGet->requestor);
	Window ManagedWindow = xObject_getWindow (toGet->requestor);

//unsure delete recieve property
	XDeleteProperty (ManagedDisplay, ManagedWindow, toGet->property);
//send convert selection request
	XConvertSelection (ManagedDisplay, toGet->selection,
			   XA_TARGETS (ManagedDisplay), toGet->property,
			   ManagedWindow, toGet->currentTime);
	while ((event.type != SelectionNotify) && (event.xselection.requestor != ManagedWindow))
		XNextEvent (toGet->requestor->xDisplay, &event);
	if (event.xselection.property != None) {
		result = XGetWindowProperty (ManagedDisplay, ManagedWindow, toGet->property, 0, 4096,	// offset - len
					     False,	// Delete
					     AnyPropertyType,	//flag
					     &type,	// return type
					     &format,	// return format
					     &len, &bytes_left,	//that 
					     &data);
		targetP = (Atom *) data;
		if (format == 32) {
			for (i = 0; i < len; i++) {
				if (*targetP == XA_UTF8_STRING (ManagedDisplay))
					UTF8_S_Enable = TRUE;
				if (*targetP == XA_STRING)
					S_Enable = TRUE;
				targetP++;
			}
		}
		else {
			send_error (WARNING, "Unknown atom format\n", NULL);
			err = WARNING;
		}

//choosing valid target. utf8_string has the highest priority.
		if (UTF8_S_Enable)
			toGet->target = XA_UTF8_STRING (ManagedDisplay);
		else {
			send_error (DEBUG, "Non-UTF8 target\n", NULL);
			if (S_Enable)
				toGet->target = XA_STRING;
			else {
				send_error (WARNING, "No valid targets found\n", NULL);
				err = ERROR;
			}
		}

	}
	else {
		send_error (WARNING, "No valid targets to convert\n", NULL);
		err = ERROR;
	}
//tell everybody, we'll finished with this property.
	XDeleteProperty (ManagedDisplay, ManagedWindow, toGet->property);
	return err;
}

status xSelection_requestToConvert (xSelection * toGet)
{
//unsure delete recieve property
	XDeleteProperty (xObject_getDisplay (toGet->requestor),
			 xObject_getWindow (toGet->requestor), toGet->property);
//send convert selection request
	toGet->target = XA_UTF8_STRING (xObject_getDisplay (toGet->requestor));

	XConvertSelection (xObject_getDisplay (toGet->requestor),
			   toGet->selection, toGet->target, toGet->property,
			   xObject_getWindow (toGet->requestor), toGet->currentTime);

	return SUCCESS;
}

XSelectionEvent *xSelection_waitForNotify (xSelection * toGet)
{
	XEvent event;
	XSelectionEvent *sevent = (XSelectionEvent *) malloc (sizeof (XSelectionEvent));
	while ((event.type != SelectionNotify)
	       && (event.xselection.requestor != xObject_getWindow (toGet->requestor)))
		XNextEvent (xObject_getDisplay (toGet->requestor), &event);
	if (event.xselection.property != None) {
		*sevent = event.xselection;
		return sevent;
	}
	else
		send_error (DEBUG, "Convert to selection target return None answer\n", NULL);

	return NULL;
}

char *xSelection_get (xSelection * toGet)
{
	char *string;
	status err = SUCCESS;
	XSelectionEvent *event;

	err = xSelection_chooseValidTarget (toGet);
	if (ERROR == err) {
		send_error (DEBUG, "Valid targets are unavailable\n", NULL);
		return NULL;
	}
	xSelection_requestToConvert (toGet);
	event = xSelection_waitForNotify (toGet);
	if (event == NULL)
		return NULL;
	string = xSelection_getToString (toGet, event);
	if (string == NULL)
		return NULL;

	free (event);

	return string;
}

status xSelection_setToNone (xSelection * toSet)
{
	XSetSelectionOwner (xObject_getDisplay (toSet->owner), toSet->selection, None, CurrentTime);
	return SUCCESS;
}

status xSelection_waitForSelectionOwnerIsAvailable (xSelection * toWait, num timeToWait)
{
	num timeout = 0;
	while (++timeout != timeToWait / MIN_CYCLE_TIME) {
		if (XGetSelectionOwner
		    (xObject_getDisplay (toWait->owner), toWait->selection) != None)
			break;
		else
			usleep (MIN_CYCLE_TIME);
	}
	if (timeout == timeToWait / MIN_CYCLE_TIME)
		return ERROR;
	else
		return SUCCESS;
}

xSwitchLang *xSwitchLang_init (Display * xDisplay)
{
	status err;
	xSwitchLang *toInit = NULL;

	toInit = (xSwitchLang *) malloc (sizeof (xSwitchLang));
	if (!toInit)
		return NULL;

	toInit->xobject = xObject_init ();
	if (!(toInit->xobject)) {
		free (toInit);
		return NULL;
	}
	toInit->xobject->xDisplay = xDisplay;
	err = XkbQueryExtension (xDisplay, NULL, NULL, NULL, NULL, NULL);
	if (!err)
		return NULL;

	XkbGetState (xDisplay, XkbUseCoreKbd, &(toInit->state));
	return toInit;
}

status xSwitchLang_free (xSwitchLang * toDelete)
{
	status err = SUCCESS;

	if (!toDelete)
		err = NULL_PTR;
	else {
		if (toDelete->xobject)
			free (toDelete->xobject);
		else
			err = NULL_PTR;
		free (toDelete);
	}

	return err;
}

//WARNING: working only with 2 languages.
status xSwitchLang_switch (xSwitchLang * toSwitch)
{
	status err = SUCCESS;

	XkbGetState (xObject_getDisplay (toSwitch->xobject), XkbUseCoreKbd, &(toSwitch->state));

	if (toSwitch->state.group == XkbGroup1Index)
		XkbLockGroup (xObject_getDisplay (toSwitch->xobject),
			      XkbUseCoreKbd, toSwitch->state.group + 1);
	else if (toSwitch->state.group == XkbGroup2Index)
		XkbLockGroup (xObject_getDisplay (toSwitch->xobject),
			      XkbUseCoreKbd, toSwitch->state.group - 1);
	else
		send_error (WARNING, "Unknown xkb group\n", NULL);
	send_error (DEBUG, "Language before switching %d\n", &(toSwitch->CurrentLang));
	if (toSwitch->CurrentLang == RUSSIAN)
		toSwitch->CurrentLang = ENGLISH;
	else
		toSwitch->CurrentLang = RUSSIAN;
	send_error (DEBUG, "Language after switching %d\n", &(toSwitch->CurrentLang));
	return err;
}

xBindingTable *xBindingTable_init (num numOfBindings)
{
	xBindingTable *toInit = NULL;

	if (numOfBindings < 1)
		return NULL;
	toInit = (xBindingTable *) malloc (sizeof (xBindingTable));
	if (toInit == NULL)
		return NULL;
	toInit->numOfBindings = numOfBindings;
	toInit->bind_list = NULL;
	toInit->bind_list = (xBinding *) malloc (numOfBindings * sizeof (xBinding));
	if (toInit->bind_list == NULL) {
		free (toInit);
		return NULL;
	}
	toInit->bind_list = memset (toInit->bind_list, 0, numOfBindings * sizeof (xBinding));
	toInit->current_action = -1;

	return toInit;
}

status xBindingTable_free (xBindingTable * toFree)
{
	status err = SUCCESS;

	if (toFree != NULL) {
		if (toFree->bind_list != NULL)
			free (toFree->bind_list);
		else
			err = WARNING;
		free (toFree);
	}
	else
		err = WARNING;

	return SUCCESS;
}

status xBindingTable_setCurrentAction (xBindingTable * table, status action)
{
	if ((action < 0) || (action > table->numOfBindings - 1))
		return INCORRECT_OPERATION;

	table->current_action = action;
	return SUCCESS;
}

status xBindingTable_getCurrentAction (xBindingTable * table)
{
	return table->current_action;
}

status
xBindingTable_bindKey (xBindingTable * table, KeySym key, KeySym key_alt,
		       KeySym modifier, status action)
{
	if ((action < 0) || (action > table->numOfBindings - 1))
		return INCORRECT_OPERATION;

	table->bind_list[action].key = key;
	table->bind_list[action].key_alt = key_alt;
	table->bind_list[action].modifier = modifier;

	return SUCCESS;
}

status xBindingTable_findBindAction (xBindingTable * table, KeySym key, KeySym modifier)
{
	num itr;
	status action = -1;

	for (itr = 0; itr < table->numOfBindings; itr++)
		if (((table->bind_list[itr].key == key)
		     && (table->bind_list[itr].modifier == modifier))
		    || ((table->bind_list[itr].key_alt == key)
			&& (table->bind_list[itr].modifier == modifier))) {
			action = itr;
			break;
		}
	return action;
}
